@@ -3,18 +3,23 @@
 
 #![deny(missing_docs)]
 
-//! This module contains the high-level interface that ties together the various
-//! engines. The main struct is [`ProcessingSession`], which knows how to run
-//! (and re-run if necessary) the various engines in the right order.
+//! The high-level Tectonic document processing interface.
 //!
-//! For an example of how to use this module, see `src/bin/tectonic.rs`, which
-//! contains tectonic's main CLI program.
+//! The main struct in this module is [`ProcessingSession`], which knows how to
+//! run (and re-run if necessary) the various engines in the right order. Such a
+//! session can be created with a [`ProcessingSessionBuilder`], which you might
+//! obtain from a [`tectonic_docmodel::document::Document`] using the
+//! [`crate::docmodel::DocumentExt::setup_session`] extension method, if you’re
+//! using the Tectonic document model. You can set one up manually if not.
+//!
+//! For an example of how to use this module, see `src/bin/tectonic/main.rs`,
+//! which contains tectonic's main CLI program.
 
 use byte_unit::Byte;
 use std::{
     collections::{HashMap, HashSet},
     fs::File,
-    io::Write,
+    io::{Read, Write},
     path::{Path, PathBuf},
     process::Command,
     rc::Rc,
@@ -22,7 +27,8 @@ use std::{
     str::FromStr,
     time::SystemTime,
 };
-use tectonic_bridge_core::{CoreBridgeLauncher, DriverHooks, SystemRequestError};
+use tectonic_bridge_core::{CoreBridgeLauncher, DriverHooks, SecuritySettings, SystemRequestError};
+use tectonic_bundles::Bundle;
 use tectonic_io_base::{
     digest::DigestData,
     filesystem::{FilesystemIo, FilesystemPrimaryInputIo},
@@ -36,7 +42,7 @@ use crate::{
     io::{
         format_cache::FormatCache,
         memory::{MemoryFileCollection, MemoryIo},
-        Bundle, InputOrigin,
+        InputOrigin,
     },
     status::StatusBackend,
     tt_error, tt_note, tt_warning,
@@ -280,6 +286,139 @@ impl BridgeState {
     fn leave_format_mode(&mut self) {
         self.format_primary = None;
     }
+
+    /// Invoke an external tool as a pass in the processing pipeline.
+    fn external_tool_pass(
+        &mut self,
+        tool: &ExternalToolPass,
+        status: &mut dyn StatusBackend,
+    ) -> Result<()> {
+        status.note_highlighted("Running external tool ", &tool.argv[0], " ...");
+
+        // Process the command arguments with our extremely-dumb fake globbing
+        // support. Filenames appearing in the arguments are treated as "read"
+        // for our rerun tracking.
+
+        let mut cmd = Command::new(&tool.argv[0]);
+        let mut read_files = HashSet::new();
+
+        {
+            let mem_files = &*self.mem.files.borrow();
+
+            for arg in &tool.argv[1..] {
+                // Primitive globbing: we replace "*.X" with matching files.
+                // This is pretty much a hack to support biber conveniently.
+                if arg.starts_with("*.") {
+                    let suffix = &arg[1..];
+
+                    for mfn in mem_files.keys() {
+                        if mfn.ends_with(suffix) {
+                            cmd.arg(mfn);
+                            read_files.insert(mfn.to_owned());
+                        }
+                    }
+                } else {
+                    cmd.arg(arg);
+
+                    if mem_files.contains_key(arg) {
+                        read_files.insert(arg.to_owned());
+                    }
+                }
+            }
+        }
+
+        // Now that we're validated, write those files to disk so that the tool
+        // can actually use them.
+
+        let tempdir = ctry!(
+            tempfile::Builder::new().tempdir();
+            "can't create temporary directory for external tool"
+        );
+
+        {
+            let mem_files = &*self.mem.files.borrow();
+
+            for name in &read_files {
+                let file = mem_files.get(name).unwrap();
+
+                let real_path = tempdir.path().join(name);
+                let mut f = ctry!(
+                    File::create(&real_path);
+                    "failed to create file `{}`", real_path.display()
+                );
+                ctry!(
+                    f.write_all(&file.data);
+                    "failed to write file `{}`", real_path.display()
+                );
+            }
+        }
+
+        // Now we can actually run the command.
+
+        let output = cmd.current_dir(tempdir.path()).output()?;
+
+        if let Some(0) = output.status.code() {
+        } else {
+            tt_error!(
+                status,
+                "the external tool exited with an error code; its stdout was:\n"
+            );
+            status.dump_error_logs(&output.stdout[..]);
+            tt_error!(status, "its stderr was:\n");
+            status.dump_error_logs(&output.stderr[..]);
+
+            return if let Some(n) = output.status.code() {
+                Err(errmsg!("the external tool exited with error code {}", n))
+            } else {
+                Err(errmsg!("the external tool was terminated by a signal"))
+            };
+        }
+
+        // Search for any files that the tool created, and import them into the
+        // memory layer.
+
+        for entry in std::fs::read_dir(tempdir.path())? {
+            let entry = entry?;
+
+            if !entry.file_type()?.is_file() {
+                continue;
+            }
+
+            if let Some(basename) = entry.file_name().to_str() {
+                if !self.mem.files.borrow().contains_key(basename) {
+                    let path = entry.path();
+                    let mut data = Vec::new();
+
+                    let mut f = ctry!(
+                        File::open(&path);
+                        "failed to open tool-created file `{}`", path.display()
+                    );
+                    ctry!(
+                        f.read_to_end(&mut data);
+                        "failed to read tool-created file `{}`", path.display()
+                    );
+
+                    self.mem.create_entry(basename, data);
+                    self.events.insert(
+                        basename.to_owned(),
+                        FileSummary::new(AccessPattern::Written, InputOrigin::NotInput),
+                    );
+                }
+            }
+        }
+
+        // Mark the input files as having been read, and we're done.
+
+        for name in &read_files {
+            let mut summ = self.events.get_mut(name).unwrap();
+            summ.access_pattern = match summ.access_pattern {
+                AccessPattern::Written => AccessPattern::WrittenThenRead,
+                c => c, // identity mapping makes sense for remaining options
+            };
+        }
+
+        Ok(())
+    }
 }
 
 macro_rules! bridgestate_ioprovider_try {
@@ -382,12 +521,24 @@ impl IoProvider for BridgeState {
         name: &str,
         status: &mut dyn StatusBackend,
     ) -> OpenResult<InputHandle> {
+        match self.input_open_name_with_abspath(name, status) {
+            OpenResult::Ok((ih, _path)) => OpenResult::Ok(ih),
+            OpenResult::Err(e) => OpenResult::Err(e),
+            OpenResult::NotAvailable => OpenResult::NotAvailable,
+        }
+    }
+
+    fn input_open_name_with_abspath(
+        &mut self,
+        name: &str,
+        status: &mut dyn StatusBackend,
+    ) -> OpenResult<(InputHandle, Option<PathBuf>)> {
         let r = (|| {
-            bridgestate_ioprovider_cascade!(self, input_open_name(name, status));
+            bridgestate_ioprovider_cascade!(self, input_open_name_with_abspath(name, status));
         })();
 
         match r {
-            OpenResult::Ok(ref ih) => {
+            OpenResult::Ok((ref ih, ref _path)) => {
                 if let Some(summ) = self.events.get_mut(name) {
                     summ.access_pattern = match summ.access_pattern {
                         AccessPattern::Written => AccessPattern::WrittenThenRead,
@@ -430,7 +581,18 @@ impl IoProvider for BridgeState {
     }
 
     fn input_open_primary(&mut self, status: &mut dyn StatusBackend) -> OpenResult<InputHandle> {
-        bridgestate_ioprovider_cascade!(self, input_open_primary(status));
+        match self.input_open_primary_with_abspath(status) {
+            OpenResult::Ok((ih, _path)) => OpenResult::Ok(ih),
+            OpenResult::Err(e) => OpenResult::Err(e),
+            OpenResult::NotAvailable => OpenResult::NotAvailable,
+        }
+    }
+
+    fn input_open_primary_with_abspath(
+        &mut self,
+        status: &mut dyn StatusBackend,
+    ) -> OpenResult<(InputHandle, Option<PathBuf>)> {
+        bridgestate_ioprovider_cascade!(self, input_open_primary_with_abspath(status));
     }
 
     fn input_open_format(
@@ -608,9 +770,44 @@ impl Default for ShellEscapeMode {
     }
 }
 
+/// A custom extra pass that invokes an external tool.
+///
+/// This is bad for reproducibility but comes in handy.
+#[derive(Debug)]
+struct ExternalToolPass {
+    argv: Vec<String>,
+}
+
+impl ExternalToolPass {
+    fn new_for_biber() -> Self {
+        // Testing support -- let the rig specify a custom executable to use,
+        // which lets us exercise different pieces of the external-tool
+        // behavior.
+
+        let s = (
+            crate::config::is_config_test_mode_activated(),
+            std::env::var("TECTONIC_TEST_FAKE_BIBER"),
+        );
+
+        let argv = match s {
+            (true, Ok(text)) => text.split_whitespace().map(|x| x.to_owned()).collect(),
+            _ => vec!["biber".to_owned(), "*.bcf".to_owned()],
+        };
+
+        ExternalToolPass { argv }
+    }
+}
+
 /// A builder-style interface for creating a [`ProcessingSession`].
+///
+/// This uses standard builder patterns. The `Default` implementation defaults
+/// to restrictive security settings that disable all known-insecure features
+/// that could be abused by untrusted inputs. Use
+/// [`ProcessingSessionBuilder::new_with_security()`] in order to have the
+/// option to enable potentially-insecure features such as shell-escape.
 #[derive(Default)]
 pub struct ProcessingSessionBuilder {
+    security: SecuritySettings,
     primary_input: PrimaryInputMode,
     tex_input_name: Option<String>,
     output_dest: OutputDestination,
@@ -633,6 +830,14 @@ pub struct ProcessingSessionBuilder {
 }
 
 impl ProcessingSessionBuilder {
+    /// Create a new builder with customized security settings.
+    pub fn new_with_security(security: SecuritySettings) -> Self {
+        ProcessingSessionBuilder {
+            security,
+            ..Default::default()
+        }
+    }
+
     /// Sets the path to the primary input file.
     ///
     /// If a primary input path is not specified, we will default to reading it from stdin.
@@ -798,7 +1003,10 @@ impl ProcessingSessionBuilder {
     /// disable shell-escape unless the [`UnstableOptions`] say otherwise,
     /// in which case a driver-managed temporary directory will be used.
     pub fn shell_escape_with_work_dir<P: AsRef<Path>>(&mut self, path: P) -> &mut Self {
-        self.shell_escape_mode = ShellEscapeMode::ExternallyManagedDir(path.as_ref().to_owned());
+        if self.security.allow_shell_escape() {
+            self.shell_escape_mode =
+                ShellEscapeMode::ExternallyManagedDir(path.as_ref().to_owned());
+        }
         self
     }
 
@@ -807,7 +1015,9 @@ impl ProcessingSessionBuilder {
     /// unless the [`UnstableOptions`] say otherwise, in which case a
     /// driver-managed temporary directory will be used.
     pub fn shell_escape_with_temp_dir(&mut self) -> &mut Self {
-        self.shell_escape_mode = ShellEscapeMode::TempDir;
+        if self.security.allow_shell_escape() {
+            self.shell_escape_mode = ShellEscapeMode::TempDir;
+        }
         self
     }
 
@@ -919,19 +1129,24 @@ impl ProcessingSessionBuilder {
         let mut pdf_path = aux_path.clone();
         pdf_path.set_extension("pdf");
 
-        let shell_escape_mode = match self.shell_escape_mode {
-            ShellEscapeMode::Defaulted => {
-                if self.unstables.shell_escape {
-                    ShellEscapeMode::TempDir
-                } else {
-                    ShellEscapeMode::Disabled
+        let shell_escape_mode = if !self.security.allow_shell_escape() {
+            ShellEscapeMode::Disabled
+        } else {
+            match self.shell_escape_mode {
+                ShellEscapeMode::Defaulted => {
+                    if self.unstables.shell_escape {
+                        ShellEscapeMode::TempDir
+                    } else {
+                        ShellEscapeMode::Disabled
+                    }
                 }
-            }
 
-            other => other,
+                other => other,
+            }
         };
 
         Ok(ProcessingSession {
+            security: self.security,
             bs,
             pass: self.pass,
             primary_input_path,
@@ -956,6 +1171,7 @@ impl ProcessingSessionBuilder {
 
 #[derive(Debug, Clone)]
 enum RerunReason {
+    Biber,
     Bibtex,
     FileChange(String),
 }
@@ -964,6 +1180,9 @@ enum RerunReason {
 /// processing a file. It understands, for example, the need to re-run the TeX
 /// engine if the `.aux` file changed.
 pub struct ProcessingSession {
+    // Security settings.
+    security: SecuritySettings,
+
     /// The subset of the session state that's can be mutated while the C/C++
     /// engines are running. Importantly, this includes the full I/O stack.
     bs: BridgeState,
@@ -1355,7 +1574,11 @@ impl ProcessingSession {
         } else {
             warnings = self.tex_pass(None, status)?;
 
-            if self.is_bibtex_needed() {
+            if self.is_biber_needed() {
+                self.bs
+                    .external_tool_pass(&ExternalToolPass::new_for_biber(), status)?;
+                Some(RerunReason::Biber)
+            } else if self.is_bibtex_needed() {
                 self.bibtex_pass(status)?;
                 Some(RerunReason::Bibtex)
             } else {
@@ -1375,6 +1598,7 @@ impl ProcessingSession {
                 "I was told to".to_owned()
             } else {
                 match rerun_result {
+                    Some(RerunReason::Biber) => "biber was run".to_owned(),
                     Some(RerunReason::Bibtex) => "bibtex was run".to_owned(),
                     Some(RerunReason::FileChange(ref s)) => format!("\"{}\" changed", s),
                     None => break,
@@ -1423,6 +1647,15 @@ impl ProcessingSession {
         Ok(0)
     }
 
+    fn is_biber_needed(&self) -> bool {
+        self.bs
+            .mem
+            .files
+            .borrow()
+            .keys()
+            .any(|filename| filename.ends_with(".bcf"))
+    }
+
     fn is_bibtex_needed(&self) -> bool {
         const BIBDATA: &[u8] = b"\\bibdata";
 
@@ -1456,7 +1689,8 @@ impl ProcessingSession {
         let result = {
             self.bs
                 .enter_format_mode(&format!("tectonic-format-{}.tex", stem));
-            let mut launcher = CoreBridgeLauncher::new(&mut self.bs, status);
+            let mut launcher =
+                CoreBridgeLauncher::new_with_security(&mut self.bs, status, self.security.clone());
             let r = TexEngine::default()
                 .halt_on_error_mode(true)
                 .initex_mode(true)
@@ -1518,7 +1752,8 @@ impl ProcessingSession {
                 status.note_highlighted("Running ", "TeX", " ...");
             }
 
-            let mut launcher = CoreBridgeLauncher::new(&mut self.bs, status);
+            let mut launcher =
+                CoreBridgeLauncher::new_with_security(&mut self.bs, status, self.security.clone());
 
             TexEngine::default()
                 .halt_on_error_mode(true)
@@ -1551,7 +1786,8 @@ impl ProcessingSession {
     fn bibtex_pass(&mut self, status: &mut dyn StatusBackend) -> Result<i32> {
         let result = {
             status.note_highlighted("Running ", "BibTeX", " ...");
-            let mut launcher = CoreBridgeLauncher::new(&mut self.bs, status);
+            let mut launcher =
+                CoreBridgeLauncher::new_with_security(&mut self.bs, status, self.security.clone());
             let mut engine = BibtexEngine::new();
             engine.process(&mut launcher, &self.tex_aux_path, &self.unstables)
         };
@@ -1583,7 +1819,8 @@ impl ProcessingSession {
         {
             status.note_highlighted("Running ", "xdvipdfmx", " ...");
 
-            let mut launcher = CoreBridgeLauncher::new(&mut self.bs, status);
+            let mut launcher =
+                CoreBridgeLauncher::new_with_security(&mut self.bs, status, self.security.clone());
             let mut engine = XdvipdfmxEngine::default();
 
             engine.build_date(self.build_date);
