@@ -1,4 +1,4 @@
-// Copyright 2016-2018 the Tectonic Project
+// Copyright 2016-2021 the Tectonic Project
 // Licensed under the MIT License.
 
 use lazy_static::lazy_static;
@@ -229,6 +229,115 @@ fn bad_outfmt_1() {
     error_or_panic(output);
 }
 
+fn run_with_biber(args: &str, stdin: &str) -> Output {
+    let fmt_arg = get_plain_format_arg();
+    let tempdir = setup_and_copy_files(&[]);
+    let mut command = prep_tectonic(tempdir.path(), &[&fmt_arg, "-"]);
+
+    let test_cmd = if cfg!(windows) {
+        format!(
+            "cmd /c {} {}",
+            util::test_path(&["fake-biber.bat"]).display(),
+            args
+        )
+    } else {
+        format!("{} {}", util::test_path(&["fake-biber.sh"]).display(), args)
+    };
+
+    command.env("TECTONIC_TEST_FAKE_BIBER", &test_cmd);
+
+    command
+        .stdin(Stdio::piped())
+        .stdout(Stdio::piped())
+        .stderr(Stdio::piped());
+    println!("running {:?}", command);
+    let mut child = command.spawn().expect("tectonic failed to start");
+
+    write!(child.stdin.as_mut().unwrap(), "{}", stdin)
+        .expect("failed to send data to tectonic subprocess");
+
+    child
+        .wait_with_output()
+        .expect("failed to wait on tectonic subprocess")
+}
+
+const BIBER_TRIGGER_TEX: &str = r#"
+% Test if we're on the second pass by seeing if the BCF already exists
+\newif\ifsecond
+\newread\rbcf
+\openin\rbcf=myfile.bcf
+\ifeof\rbcf
+\message{first pass}
+\secondfalse
+\else
+\message{second pass}
+\secondtrue
+\closein\rbcf
+\fi
+
+% Now create BCF
+\newwrite\wbcf
+\immediate\openout\wbcf=myfile.bcf\relax
+\immediate\write\wbcf{Hello BCF file}
+\immediate\closeout\wbcf
+"#;
+
+#[test]
+fn biber_failure() {
+    let output = run_with_biber("failure", BIBER_TRIGGER_TEX);
+    error_or_panic(output);
+}
+
+#[test]
+fn biber_no_such_tool() {
+    let fmt_arg = get_plain_format_arg();
+    let tempdir = setup_and_copy_files(&[]);
+    let mut command = prep_tectonic(tempdir.path(), &[&fmt_arg, "-"]);
+
+    command.env("TECTONIC_TEST_FAKE_BIBER", "ohnothereisnobiberprogram");
+
+    const REST: &str = r#"\bye"#;
+    let tex = format!("{}{}", BIBER_TRIGGER_TEX, REST);
+
+    command
+        .stdin(Stdio::piped())
+        .stdout(Stdio::piped())
+        .stderr(Stdio::piped());
+    println!("running {:?}", command);
+    let mut child = command.spawn().expect("tectonic failed to start");
+
+    write!(child.stdin.as_mut().unwrap(), "{}", tex)
+        .expect("failed to send data to tectonic subprocess");
+
+    let output = child
+        .wait_with_output()
+        .expect("failed to wait on tectonic subprocess");
+    error_or_panic(output);
+}
+
+#[cfg(unix)]
+#[test]
+fn biber_signal() {
+    let output = run_with_biber("signal", BIBER_TRIGGER_TEX);
+    error_or_panic(output);
+}
+
+#[test]
+fn biber_success() {
+    const REST: &str = r#"
+\ifsecond
+\ifnum\input{biberout.qqq}=456\relax
+a
+\else
+\ohnothebiberdidntwork
+\fi
+\fi
+\bye"#;
+    let tex = format!("{}{}", BIBER_TRIGGER_TEX, REST);
+    let output = run_with_biber("success", &tex);
+    success_or_panic(output);
+}
+
 #[test]
 fn help_flag() {
     let output = run_tectonic(&PathBuf::from("."), &["-h"]);
@@ -567,6 +676,16 @@ fn v2_new_build_multiple_outputs() {
     success_or_panic(output);
 }
 
+const SHELL_ESCAPE_TEST_DOC: &str = r#"\immediate\write18{mkdir shellwork}
+\immediate\write18{echo 123 >shellwork/persist}
+\ifnum123=\input{shellwork/persist}
+a
+\else
+\ohnotheshellescapedidntwork
+\fi
+\bye
+"#;
+
 /// Test that shell escape actually runs the commands
 #[test]
 fn shell_escape() {
@@ -576,15 +695,49 @@ fn shell_escape() {
     let output = run_tectonic_with_stdin(
         tempdir.path(),
         &[&fmt_arg, "-", "-Zshell-escape"],
-        r#"\immediate\write18{mkdir shellwork}
-        \immediate\write18{echo 123 >shellwork/persist}
-        \ifnum123=\input{shellwork/persist}
-        a
-        \else
-        \ohnotheshellescapedidntwork
-        \fi
-        \bye
-        "#,
+        SHELL_ESCAPE_TEST_DOC,
     );
     success_or_panic(output);
 }
+
+/// Test that shell-escape can be killed by command-line-option
+#[test]
+fn shell_escape_cli_override() {
+    let fmt_arg = get_plain_format_arg();
+    let tempdir = setup_and_copy_files(&[]);
+
+    let output = run_tectonic_with_stdin(
+        tempdir.path(),
+        &[&fmt_arg, "--untrusted", "-", "-Zshell-escape"],
+        SHELL_ESCAPE_TEST_DOC,
+    );
+    error_or_panic(output);
+}
+
+/// Test that shell-escape can be killed by environment variable
+#[test]
+fn shell_escape_env_override() {
+    let fmt_arg = get_plain_format_arg();
+    let tempdir = setup_and_copy_files(&[]);
+
+    // Note that we intentionally set the variable to 0 below -- it takes it
+    // effect if it has ANY value, not just a "truthy" one.
+
+    let mut command = prep_tectonic(tempdir.path(), &[&fmt_arg, "-", "-Zshell-escape"]);
+    command
+        .stdin(Stdio::piped())
+        .stdout(Stdio::piped())
+        .stderr(Stdio::piped())
+        .env("TECTONIC_UNTRUSTED_MODE", "0");
+
+    println!("running {:?}", command);
+    let mut child = command.spawn().expect("tectonic failed to start");
+    write!(child.stdin.as_mut().unwrap(), "{}", SHELL_ESCAPE_TEST_DOC)
+        .expect("failed to send data to tectonic subprocess");
+
+    let output = child
+        .wait_with_output()
+        .expect("failed to wait on tectonic subprocess");
+
+    error_or_panic(output);
+}
